﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2017 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.Algorithms.DataAnalysis {
  [StorableClass]
  internal class M5RuleModel : RegressionModel {
    #region Properties
    [Storable]
    internal string[] SplitAttributes { get; private set; }
    [Storable]
    private double[] splitValues;
    [Storable]
    private Comparison[] comparisons;
    [Storable]
    protected IRegressionModel RuleModel { get; set; }
    [Storable]
    private IReadOnlyList<string> variables;
    #endregion

    #region HLConstructors
    [StorableConstructor]
    protected M5RuleModel(bool deserializing) : base(deserializing) { }
    protected M5RuleModel(M5RuleModel original, Cloner cloner) : base(original, cloner) {
      if (original.SplitAttributes != null) SplitAttributes = original.SplitAttributes.ToArray();
      if (original.splitValues != null) splitValues = original.splitValues.ToArray();
      if (original.comparisons != null) comparisons = original.comparisons.ToArray();
      RuleModel = cloner.Clone(original.RuleModel);
      if (original.variables != null) variables = original.variables.ToList();
    }
    private M5RuleModel(string target) : base(target) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new M5RuleModel(this, cloner);
    }
    #endregion

    internal static M5RuleModel CreateRuleModel(string target, M5Parameters m5Params) {
      return m5Params.LeafModel.ProvidesConfidence ? new ConfidenceM5RuleModel(target) : new M5RuleModel(target);
    }

    #region IRegressionModel
    public override IEnumerable<string> VariablesUsedForPrediction {
      get { return variables; }
    }

    public override IEnumerable<double> GetEstimatedValues(IDataset dataset, IEnumerable<int> rows) {
      if (RuleModel == null) throw new NotSupportedException("The model has not been built correctly");
      return RuleModel.GetEstimatedValues(dataset, rows);
    }

    public override IRegressionSolution CreateRegressionSolution(IRegressionProblemData problemData) {
      return new RegressionSolution(this, problemData);
    }
    #endregion


    public void Build(IReadOnlyList<int> trainingRows, IReadOnlyList<int> pruningRows, M5Parameters m5Params, CancellationToken cancellationToken) {
      variables = m5Params.AllowedInputVariables.ToList();
      var tree = M5TreeModel.CreateTreeModel(m5Params.TargetVariable, m5Params);
      tree.Build(trainingRows, pruningRows, m5Params, cancellationToken);
      var nodeModel = tree.Root.EnumerateNodes().Where(x => x.IsLeaf).MaxItems(x => x.NumSamples).First();

      var satts = new List<string>();
      var svals = new List<double>();
      var reops = new List<Comparison>();

      //extract Splits
      for (var temp = nodeModel; temp.Parent != null; temp = temp.Parent) {
        satts.Add(temp.Parent.SplitAttribute);
        svals.Add(temp.Parent.SplitValue);
        reops.Add(temp.Parent.Left == temp ? Comparison.LessEqual : Comparison.Greater);
      }
      RuleModel = nodeModel.Model;
      comparisons = reops.ToArray();
      SplitAttributes = satts.ToArray();
      splitValues = svals.ToArray();
    }

    public void Update(IReadOnlyList<int> rows, M5Parameters m5Parameters, CancellationToken cancellationToken) {
      BuildModel(rows, m5Parameters.Random, m5Parameters.Data, m5Parameters.LeafModel, cancellationToken);
    }

    public bool Covers(IDataset dataset, int row) {
      return !SplitAttributes.Where((t, i) => !comparisons[i].Compare(dataset.GetDoubleValue(t, row), splitValues[i])).Any();
    }

    public string ToCompactString() {
      var mins = new Dictionary<string, double>();
      var maxs = new Dictionary<string, double>();
      for (var i = 0; i < SplitAttributes.Length; i++) {
        var n = SplitAttributes[i];
        var v = splitValues[i];
        if (!mins.ContainsKey(n)) mins.Add(n, double.NegativeInfinity);
        if (!maxs.ContainsKey(n)) maxs.Add(n, double.PositiveInfinity);
        if (comparisons[i] == Comparison.LessEqual) maxs[n] = Math.Min(maxs[n], v);
        else mins[n] = Math.Max(mins[n], v);
      }
      if (maxs.Count == 0) return "";
      var s = new StringBuilder();
      foreach (var key in maxs.Keys)
        s.Append(string.Format("{0} ∈ [{1:e2}; {2:e2}] && ", key, mins[key], maxs[key]));
      s.Remove(s.Length - 4, 4);
      return s.ToString();
    }

    #region Helpers
    private void BuildModel(IReadOnlyList<int> rows, IRandom random, IDataset data, ILeafModel leafModel, CancellationToken cancellationToken) {
      var reducedData = new Dataset(VariablesUsedForPrediction.Concat(new[] {TargetVariable}), VariablesUsedForPrediction.Concat(new[] {TargetVariable}).Select(x => data.GetDoubleValues(x, rows).ToList()));
      var pd = new RegressionProblemData(reducedData, VariablesUsedForPrediction, TargetVariable);
      pd.TrainingPartition.Start = 0;
      pd.TrainingPartition.End = pd.TestPartition.Start = pd.TestPartition.End = reducedData.Rows;

      int noparams;
      RuleModel = leafModel.Build(pd, random, cancellationToken, out noparams);
      cancellationToken.ThrowIfCancellationRequested();
    }
    #endregion

    [StorableClass]
    private sealed class ConfidenceM5RuleModel : M5RuleModel, IConfidenceRegressionModel {
      #region HLConstructors
      [StorableConstructor]
      private ConfidenceM5RuleModel(bool deserializing) : base(deserializing) { }
      private ConfidenceM5RuleModel(ConfidenceM5RuleModel original, Cloner cloner) : base(original, cloner) { }
      public ConfidenceM5RuleModel(string targetAttr) : base(targetAttr) { }
      public override IDeepCloneable Clone(Cloner cloner) {
        return new ConfidenceM5RuleModel(this, cloner);
      }
      #endregion

      public IEnumerable<double> GetEstimatedVariances(IDataset dataset, IEnumerable<int> rows) {
        return ((IConfidenceRegressionModel)RuleModel).GetEstimatedVariances(dataset, rows);
      }

      public override IRegressionSolution CreateRegressionSolution(IRegressionProblemData problemData) {
        return new ConfidenceRegressionSolution(this, problemData);
      }
    }
  }

  internal enum Comparison {
    LessEqual,
    Greater
  }

  internal static class ComparisonExtentions {
    public static bool Compare(this Comparison op, double x, double y) {
      switch (op) {
        case Comparison.Greater:
          return x > y;
        case Comparison.LessEqual:
          return x <= y;
        default:
          throw new ArgumentOutOfRangeException(op.ToString(), op, null);
      }
    }
  }
}